package salsa_lite.core.language;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.net.InetAddress;
import java.lang.reflect.Method;
import java.lang.reflect.InvocationTargetException;
import java.util.Vector;
import java.util.Hashtable;

import salsa_lite.core.language.exceptions.CurrentContinuationException;
import salsa_lite.core.language.exceptions.MessageHandlerNotFoundException;
import salsa_lite.core.naming.UAN;
import salsa_lite.core.naming.UAL;
import salsa_lite.core.naming.URI;
import salsa_lite.core.services.ErrorService;
import salsa_lite.core.services.OutputService;
import salsa_lite.core.services.ServiceFactory;

public class LiteActorState implements Actor, java.io.Serializable {
	public transient OutputService standardOutput;
	public transient ErrorService standardError;

	private UAN uan;
	private UAL ual;
	private int hashcode;

	public UAN getUAN()		{ return uan; }
	public UAL getUAL()		{ return ual; }
	public URI getURI() {
		if (uan != null) return uan;
		else return ual;
	}


	public int hashCode()		{ return hashcode; }

	public String getUniqueIdentifier() {
		if (uan != null) {
			return uan.toString();
		} else if (ual != null) {
			return ual.toString();
		} else {
			System.err.println("ActorReference error:");
			System.err.println("\tUnidentified Actor Reference.");
			return null;
		}
	}

	public boolean equals(Object o) {
		if (o instanceof Actor) {
			Actor other = (Actor)o;
			return getUniqueIdentifier().equals(other.getUniqueIdentifier());
		} else if (o instanceof ActorReference) {
			ActorReference other = (ActorReference)o;
			return getUniqueIdentifier().equals(other.getUniqueIdentifier());
		}
		return false;
	}

	public String toString() {
		return "[" + getClass().getName() + ", " + getUniqueIdentifier() + "]";
	}

	/********
		*	Constructors for LiteActor.State
		*	State(uan,host,port) is called when an actor is created through the construct or main method
		*	State(String identifier) is called by the service factory for system actors
	 ********/
	public ActorReference self;
	public ActorReference getReference() { return self; }

	public LiteActorState(UAN uan) {
		this.uan = uan;
		self = new LiteActor(this.uan);

		standardOutput = ServiceFactory.getOutput();
		standardError = ServiceFactory.getError();
		hashcode = getUniqueIdentifier().hashCode();
	}

	public LiteActorState(UAL ual) {
		this.ual = ual;
		self = new LiteActor(this.ual);

		standardOutput = ServiceFactory.getOutput();
		standardError = ServiceFactory.getError();
		hashcode = getUniqueIdentifier().hashCode();
	}

	/********
		*	Only system actors are created with this method.
	 ********/
	public LiteActorState(String identifier) {
		try {
			this.ual = new UAL("rmsp://" + ServiceFactory.getReception().getLocalHost() + ":" + ServiceFactory.getReception().getLocalPort() + "/" + identifier);
		} catch (Exception e) {
			System.err.println("ERROR: generating UAL for system actor:" + e);
			e.printStackTrace();
			return;
		}
		hashcode = getUniqueIdentifier().hashCode();
	}

	private Message constructMessage;
	public void setConstructMessage(Message constructMessage) {
		this.constructMessage = constructMessage;
	}
	public Message getConstructMessage() {
		Message retval = constructMessage;
		constructMessage = null;
		return retval;
	}



	public Message currentMessage = null;
	public Vector<Message> __messages = new Vector<Message>();
	public Vector<Message> getGeneratedMessages() {
		Vector<Message> messages = __messages;
		__messages = new Vector<Message>();
		return messages;
	}

	public void setCurrentMessage(Message message) {
		currentMessage = message;
	}

	/********
		*	For garbage collection
	 ********/
	private boolean destroyed = false;
	public boolean isDestroyed() { return destroyed; }
	public void destroy() {
		if (uan != null) {
			ServiceFactory.getNaming().delete(uan);
		}
		destroyed = true;
	}

//	protected void finalize() {
//		System.err.println("COLLECTING: " + getURI().getIdentifier());
//	}

	/********
		*	For Migration
	 ********/
	private boolean migrating = false;
	private String migrationHost = null;
	private int migrationPort = -1;
	private Message migrationMessage = null;

	public boolean	isMigrating()			{ return migrating; }
	public String	getMigrationHost()		{ return migrationHost; }
	public int	getMigrationPort()		{ return migrationPort; }

	public void migrate(String host, int port) {
		if (getUAN() == null) {
			System.err.println("Migration ERROR: cannot migrate actor without UAN: " + this);
			return;
		}

		try {
			migrationHost = InetAddress.getByName(host).getHostAddress();
		} catch (Exception e) {
			System.err.println("Migrate Error: unknown host for actor: " + e);
			e.printStackTrace();
			return;
		}
		migrationPort = port;

		if (ServiceFactory.getReception().isLocal(migrationHost, migrationPort)) {
			/********
				*	Actor is already at the migration target, do nothing.
			 ********/
			migrationHost = null;
			migrationPort = -1;
			return;
		} else {
			migrating = true;
			migrationMessage = currentMessage;
		}

		try {
			ServiceFactory.getNaming().update(uan, new UAL("rmsp://" + migrationHost + ":" + migrationPort + uan.getIdentifier()));
		} catch (Exception e) {
			System.err.println("Migrate Error: generating UAL from UAN:" + e);
			e.printStackTrace();
		}
	}

	public void migrationCompleted() {
		migrating = false;
		migrationHost = null;
		migrationPort = -1;

		standardOutput = ServiceFactory.getOutput();
		standardError = ServiceFactory.getError();

		if (migrationMessage != null) {
			migrationMessage.resolveContinuations(null);
			migrationMessage = null;
		}
	}

	/********
		*	The following handles token passing.
	 ********/
	public Hashtable<String, Message> pendingMessages = new Hashtable<String, Message>();
	public Vector<UnresolvedTokenWrapper> unresolvedTokens = new Vector<UnresolvedTokenWrapper>();

	/**
	 * This class wraps the information from a resolveToken message,
	 * into an object so only one Vector is needed to store unresolved
	 * tokens.
	 */
	private class UnresolvedTokenWrapper implements java.io.Serializable {
		protected String        messageId;
		protected int           position;
		protected Object        value;

		public UnresolvedTokenWrapper(String messageId, int position, Object value) {
			this.messageId = messageId;
			this.position = position;
			this.value = value;
		}
	}

	public void handleRequiredTokens(Message message) {
		String messageId = message.getMessageId();
		pendingMessages.put(messageId, message);

		/**
		 * We need to check and see if any tokens that were meant for
		 * this message have already been receieved.  This requires
		 * going through the vector of unresolved tokens, and resolving
		 * the ones intended for this message.
		 */
		for (int i = unresolvedTokens.size()-1; i >= 0; i--) {
			UnresolvedTokenWrapper unresolvedToken = unresolvedTokens.get(i);

			if (unresolvedToken.messageId.equals(messageId)) {
				resolveToken(messageId, unresolvedToken.position, unresolvedToken.value);
				unresolvedTokens.remove(i);
			}
		}
	}

	/**
	 * resolveToken messages are sent when a message has completed processing, and
	 * returns a token that other messages are waiting for.
	 * This method updates the specified message with the given value, if the message
	 * has already been received, if not the information is placed in an
	 * UnresolvedTokenWrapper and put in a Vector of unresolved tokens.
	 */
	public void resolveToken(String messageId, int position, Object value) {
		/**
		 *      Try and resolve this token, if no matches are found
		 *      add it to the unresolvedHashtables because its
		 *      target message hasn't reached this actor yet.
		 */
		Message targetMessage = (Message)pendingMessages.get(messageId);

		if (targetMessage != null) {
			targetMessage.resolveToken(value, position);

			if (targetMessage.getRequiredTokens() == 0) {
				pendingMessages.remove(messageId);
				ServiceFactory.getStage().sendMessageFromLocal(targetMessage);
			}
		} else {
			unresolvedTokens.add( new UnresolvedTokenWrapper(messageId, position, value) );
		}
	}

	/********
		*	The following provides support message invocation.
	 ********/
	public Object invokeMessage(String methodName, Object[] arguments) throws CurrentContinuationException, MessageHandlerNotFoundException, Exception {
		if (methodName.equals("resolveToken")) {
			resolveToken((String)arguments[0], ((Integer)arguments[1]).intValue(), arguments[2]);
			return null;
		} else if (methodName.equals("migrate")) {
			migrate((String)arguments[0], ((Integer)arguments[1]).intValue());
			return null;
		} else if (methodName.equals("toString")) {
			return toString();
		} else if (methodName.equals("getUniqueIdentifier")) {
			return getUniqueIdentifier();
		} else if (methodName.equals("getUAN")) {
			return getUAN();
		} else if (methodName.equals("getUAL")) {
			return getUAL();
		} else if (methodName.equals("getClass")) {
			return getClass();
		} else if (methodName.equals("destroy")) {
			destroy();
			return null;
		} else {
			throw new MessageHandlerNotFoundException(methodName, arguments);
		}
	}
}
