package salsa_lite.local;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.lang.reflect.Method;
import java.lang.reflect.InvocationTargetException;
import java.util.Vector;
import java.util.Hashtable;

import salsa_lite.core.language.Actor;
import salsa_lite.core.language.ActorReference;
import salsa_lite.core.language.Message;
import salsa_lite.core.language.exceptions.CurrentContinuationException;
import salsa_lite.core.naming.UAN;
import salsa_lite.core.naming.UAL;
import salsa_lite.core.services.StageService;
import salsa_lite.core.services.ServiceFactory;

public class StageActor extends Thread {

	private StageService stageService;

	private Hashtable<String,Actor> actors;
	private Vector<Message> mailbox;
	private int id;
	
	public StageActor(int id) {
		this.id = id;
		actors = new Hashtable<String,Actor>();
		mailbox = new Vector<Message>();
	}

	public synchronized void putMessageInMailbox(Message message) {
		if (message.getProperty() != null && message.getProperty().equals("priority")) {
			mailbox.insertElementAt(message, 0);
		} else {
			mailbox.add(message);
		}
		notify();
	}

	private synchronized Message getMessage() {
		if (mailbox.isEmpty()) {
			try {
				wait();
			} catch (InterruptedException e) {
				System.err.println("Stage Error:");
				System.err.println("\tInterruptedException in getMessage(): " + e);
				System.exit(0);
			}
		}
		return mailbox.remove(0);
	}

	public void putActorOnStage(Actor actor) {
		String unique_identifier = actor.getUniqueIdentifier();

		synchronized(actors) {
			actors.put(unique_identifier, actor);
		}
		Message constructMessage = actor.getConstructMessage();
		if (constructMessage != null) process(actor, constructMessage);
	}

	public void run() {
		stageService = ServiceFactory.getStage();

		Message currentMessage;
		String targetID;
		Actor currentTarget;

		while (true) {
			currentMessage = getMessage();
			targetID = currentMessage.getTarget().getUniqueIdentifier();

			synchronized(actors) {
				currentTarget = actors.get(targetID);
			}

			if (currentTarget != null) {
				process(currentTarget, currentMessage);
				if (currentTarget.isDestroyed()) {
					actors.remove(targetID);
				}
			} else {
				/********
					*	Could not find the target actor at this stage,
					*	have the stage service determine where to send
					*	the message.
				 ********/
				System.err.println("couldn't find target for message: " + currentMessage);
			}
		}
	}

	/********
		*	The following handles message lookup and invocation.
	 ********/
	private void process(Actor actor, Message message) {
		if (message.getRequiredTokens() > 0) {
			actor.handleRequiredTokens(message);
			return;
		}

		actor.setCurrentMessage(message);
		try {
			Object returnValue = actor.invokeMessage(message.getMethodName(), message.getArguments());
			message.resolveContinuations(returnValue);
		} catch (CurrentContinuationException e) {
		} catch (Exception exception) {
			String exceptionMessage = "Message processing exception:\n";
			exceptionMessage += "\tSent by: " + message.getSource() + "\n";
			exceptionMessage += "\tReceived by actor: " + actor.toString() + "\n";
			exceptionMessage += "\tMessage: " + message.toString() + "\n";
			exceptionMessage += "\tThrew exception: " + exception + "\n";
			System.err.println(exceptionMessage);
			exception.printStackTrace();
			return;
		}
		Vector<Message> generatedMessages = actor.getGeneratedMessages();
		for (int i = 0; i < generatedMessages.size(); i++) stageService.sendMessageFromLocal( generatedMessages.get(i) );
		actor.setCurrentMessage(null);
	}
}
