/** SALSA/World Wide Computer Project
 *  ObjectHandler recieves incoming objects
 *  and determines if they are actors or messages.
 *  upon recieving actors, it will start them up.
 *  messages get sent to the appropriate actor.
 *
 *  author Travis Desell
 */
package salsa_lite.wwc.reception;

import java.net.InetAddress;
import java.net.Socket;
import java.net.ServerSocket;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.io.ObjectInputStream;

import salsa_lite.core.language.Actor;
import salsa_lite.core.language.Message;

import salsa_lite.core.services.ReceptionService;
import salsa_lite.core.services.ServiceFactory;
import salsa_lite.core.services.StageService;

public class WWCReceptionService extends Thread implements ReceptionService {
	private ServerSocket server = null;
	private String host;
	private int port;

	public String getLocalHost()		{ return host; }
	public int getLocalPort()		{ return port; }

	public boolean isLocal(String host, int port) {
		return this.host.equals(host) && this.port == port;
	}

	public void handleIncomingSocket(Socket incoming_socket) {
		final Socket incoming = incoming_socket;
		Runnable incomingSocketHandler = new Runnable() {
			public void run() {
				StageService stageService = ServiceFactory.getStage();
				String remoteHost;
				int remotePort;
				Object received = null;
				ObjectInputStream stream = null;

				try {
					stream = new ObjectInputStream(incoming.getInputStream());
				} catch (IOException e) {
					System.err.println("Reception Service error: ");
					System.err.println("\tIOException occured creating ObjectInputStream from remote host.");
					System.err.println("\tException: " + e);
					System.err.println("\tException Message: " + e.getMessage());
					e.printStackTrace();
					return;
				}

				try {
					remoteHost = (String)stream.readObject();
					remotePort = (Integer)stream.readObject();
				} catch (Exception e) {
					System.err.println("Reception Service error: ");
					System.err.println("\tException occured reading remote host and port.");
					System.err.println("\tException: " + e);
					System.err.println("\tException Message: " + e.getMessage());
					e.printStackTrace();
					return;
				}

				if (ServiceFactory.getReception().isLocal(remoteHost, remotePort)) {
					System.err.println("IncomingSocketHandler ERROR: received incoming connection from same theater.");
					System.err.println("\thost:port:: " + remoteHost + ":" + remotePort);
				}

				while (true) {
					try {
						received = stream.readObject();
					} catch (ClassCastException e) {
						System.err.println("Reception Service error: ");
						System.err.println("\tIncomingSocketHandler received an unknown object.");
						System.err.println("\treceived: " + received);
						e.printStackTrace();
					} catch (ClassNotFoundException e) {
						System.err.println("Reception Service error: ");
						System.err.println("\tCould not load class for an incoming object");
						System.err.println("\tIs it in the theater CLASSPATH?");
						System.err.println("\tException: " + e);
						System.err.println("\treceived: " + received);
						e.printStackTrace();
						return;
					} catch (IOException e) {
						System.err.println("Reception Service error: ");
						System.err.println("\tIOException occured closing connection.");
						System.err.println("\tException: " + e);
						System.err.println("\treceived: " + received);
						e.printStackTrace();
						return;
					}
	
					if (received instanceof Actor) {
						stageService.actorMigration((Actor)received, remoteHost, remotePort);
					} else if (received instanceof Message) {
						stageService.sendMessageFromRemote((Message)received, remoteHost, remotePort);
					} else {
						stageService.handleObjectFromRemote(received, remoteHost, remotePort);
					}
				}
			}
		};
		new Thread (incomingSocketHandler, "IncomingSocketHandler").start();
	}

	public WWCReceptionService() {
		String portProperty = System.getProperty("port");
		if (portProperty == null) port = 4040;
		else port = Integer.parseInt(portProperty);

		try {
			host = InetAddress.getLocalHost().getHostAddress();
		} catch (Exception e) {
			System.err.println("Error determining localhost: " + e);
			e.printStackTrace();
			return;
		}
		try {
			server = new ServerSocket( port );
		} catch (IOException e) {
			System.err.println("Theater Service error:");
			System.err.println("\tCould not start theater.");
			System.err.println("\tException: " + e);
			return;
		}
		// Check to see what the port actually is.
		port = server.getLocalPort();


		System.out.println("WWCReceptionService listening on: " + host + ":" + port);

		start();
	}

	public void run() {
		while (true) {
			try {
				Socket incoming = server.accept();
				handleIncomingSocket(incoming);
			} catch (IOException e) {
				System.err.println("Reception Service error: ");
				System.err.println("\tFailed to accept connection.");
				System.err.println("\teException: " + e);
			}
		}
	}
}

