package salsa_lite.wwc.stage;

import java.util.Hashtable;
import java.util.Vector;

import salsa_lite.core.language.Actor;
import salsa_lite.core.language.ActorReference;
import salsa_lite.core.language.Message;
import salsa_lite.core.language.exceptions.CurrentContinuationException;
import salsa_lite.core.naming.UAN;
import salsa_lite.core.naming.UAL;
import salsa_lite.core.naming.URI;
import salsa_lite.core.services.NamingService;
import salsa_lite.core.services.ReceptionService;
import salsa_lite.core.services.ServiceFactory;
import salsa_lite.core.services.StageService;
import salsa_lite.core.services.TransportService;


public class StageActor extends Thread {

	private int stage_id;
	private Vector<Object> mailbox;
	private Vector<String> alreadyUpdating;
	private Hashtable<String, Actor> actors;
	private Hashtable<String, Integer> actorStates;
	private Hashtable<String, String> hosts;
	private Hashtable<String, Integer> ports;
	private Hashtable<String, Vector<Object>> pendingMessages;

	private NamingService namingService;
	private ReceptionService receptionService;
	private StageService stageService;
	private TransportService transportService;

	protected StageActor(int stage_id) {
		this.stage_id = stage_id;

		mailbox = new Vector<Object>();
		alreadyUpdating = new Vector<String>();

		actors = new Hashtable<String, Actor>();
		actorStates = new Hashtable<String, Integer>();
		hosts = new Hashtable<String, String>();
		ports = new Hashtable<String, Integer>();
		pendingMessages = new Hashtable<String, Vector<Object>>();
	}

	public void run() {
		setName("Stage_" + stage_id);
		System.err.println("STARTING STAGE: " + getName());

		namingService = ServiceFactory.getNaming();
		receptionService = ServiceFactory.getReception();
		stageService = ServiceFactory.getStage();
		transportService = ServiceFactory.getTransport();

		while (true) {
			Object current = getNext();

			if (current instanceof SendFromLocalMessage)			handleSendFromLocal( (SendFromLocalMessage)current );
			else if (current instanceof CreateActorMessage)			handleCreateActor( (CreateActorMessage)current );
			else if (current instanceof SendFromRemoteMessage)		handleSendFromRemote( (SendFromRemoteMessage)current );
			else if (current instanceof ActorMigrationMessage)		handleActorMigration( (ActorMigrationMessage)current );
			else if (current instanceof UpdateInvalidReferenceMessage)	handleUpdateInvalidReferenceMessage( (UpdateInvalidReferenceMessage) current );
			else if (current instanceof UpdatedInvalidReferenceMessage)	handleUpdatedInvalidReferenceMessage( (UpdatedInvalidReferenceMessage) current );
			else {
				System.err.println("Received unknown object: " + current);
			}
		}
	}

	/********
		*	The following handles a message sent by an actor at this theater.
	 ********/
	public void handleSendFromLocal(SendFromLocalMessage sflm) {
		Message message = sflm.getMessage();
		URI targetURI = message.getTargetURI();
		String targetId = message.getTargetIdentifier();
		Actor targetActor = actors.get(targetId);

		/********
			*	If the actor is in the actors hashtable then it is local and
			*	process the message. Otherwise determine if the target is
			*	in transit, remote, or invalid.
		 ********/
		if (targetActor != null) {
			process(targetActor, message);
		} else {
			String targetHost = hosts.get(targetId);
			int targetPort;
			if (targetHost == null) {
				/********
					*	The location of the actor is unknown, if the URI is a UAN, then
					*	do a lookup to determine it's host and port.  If the URI is a
					*	UAL and local, then the UAL is invalid.
				 ********/
				if (targetURI instanceof UAN) {
					UAL resolved = namingService.get( (UAN)targetURI );
					targetHost = resolved.getHost();
					targetPort = resolved.getPort();
					hosts.put(targetId, targetHost);
					ports.put(targetId, targetPort);
				} else {
					targetHost = targetURI.getHost();
					targetPort = targetURI.getPort();

					if (receptionService.isLocal(targetHost, targetPort)) {
						System.err.println("STAGE ERROR: could not send message because identifier was invalid: " + targetURI);
						System.err.println("\tmessage: " + message);
						return;
					}

					hosts.put(targetId, targetHost);
					ports.put(targetId, targetPort);
				}
			} else {
				targetPort = ports.get(targetId);
			}

			/********
				*	The location of the actor is known.  if it is remote, then send
				*	the message to that theater.  if it is local, then the actor is
				*	in transit to this theater.
			 ********/
			if (receptionService.isLocal(targetHost, targetPort)) {
				Vector<Object> storedMessages = pendingMessages.get(targetId);
				if (storedMessages == null) storedMessages = new Vector<Object>();
				storedMessages.add(sflm);
				pendingMessages.put(targetId, storedMessages);
			} else {
				transportService.transfer(message, targetHost, targetPort);
			}
		}
	}

	/********
		*	The following handles a message sent by an actor from a remote theater.
	 ********/
	public void handleSendFromRemote(SendFromRemoteMessage sfrm) {
		Message message = sfrm.getMessage();
		URI targetURI = message.getTargetURI();
		String targetId = message.getTargetIdentifier();
		Actor targetActor = actors.get(targetId);

		/********
			*	If the actor is in the actors hashtable then it is local and
			*	process the message. Otherwise determine if the target is
			*	in transit, remote, or invalid.
		 ********/
		if (targetActor != null) {
			process(targetActor, message);
		} else {
			String targetHost = hosts.get(targetId);
			int targetPort;
			if (targetHost == null) {
				/********
					*	The location of the actor is unknown, if the URI is a UAN, then
					*	do a lookup to determine it's host and port. If the URI is a UAL 
					*	and not local then it is an invalid URI.
				 ********/
				if (targetURI instanceof UAN) {
					UAL resolved = namingService.get( (UAN)targetURI );
					targetHost = resolved.getHost();
					targetPort = resolved.getPort();
					hosts.put(targetId, targetHost);
					ports.put(targetId, targetPort);
				} else {
					targetHost = targetURI.getHost();
					targetPort = targetURI.getPort();

					if (!receptionService.isLocal(targetHost, targetPort)) {
						System.err.println("STAGE ERROR: could not send message because identifier was invalid: " + targetURI);
						System.err.println("\tmessage: " + message);
						return;
					}

					hosts.put(targetId, targetHost);
					ports.put(targetId, targetPort);
				}
			} else {
				targetPort = ports.get(targetId);
			}

			/********
				*	The location of the actor is known.  if it is remote, then send
				*	the message to that theater.  if it is local, then the actor is
				*	in transit to this theater.
			 ********/
			if (receptionService.isLocal(targetHost, targetPort)) {
				Vector<Object> storedMessages = pendingMessages.get(targetId);
				if (storedMessages == null) storedMessages = new Vector<Object>();
				storedMessages.add(sfrm);
				pendingMessages.put(targetId, storedMessages);
			} else {
				if (targetURI instanceof UAN) {
					sendUpdateMessage(sfrm);
				} else {
					System.err.println("SHOULDN'T REACH THIS!");
					System.exit(0);
				}
				transportService.transfer(message, targetHost, targetPort);
			}
		}
	}

	/********
		*	The following handles incoming actors from remote theaters.
	 ********/
	private void handleActorMigration(ActorMigrationMessage amm) {
		Actor actor = amm.getActor();
		String id = actor.getUniqueIdentifier();

		actors.put(id, actor);
		hosts.remove(id);
		ports.remove(id);

		actor.migrationCompleted();
		Message constructMessage = actor.getConstructMessage();
		if (constructMessage != null) process(actor, constructMessage);

		Vector<Object> storedMessages = pendingMessages.remove(id);
		if (storedMessages != null) {
			for (int i = 0; i < storedMessages.size(); i++) {
				Object current = storedMessages.get(i);
				if (current instanceof SendFromLocalMessage) put( (SendFromLocalMessage)current );
				else if (current instanceof SendFromRemoteMessage) put( (SendFromRemoteMessage)current );
				else put(current);
			}
		}
	}

	/********
		*	The following handles actors created at this theater (which may need to be created remotely).
	 ********/
	private void handleCreateActor(CreateActorMessage cam) {
		Actor actor = cam.getActor();
		String id = actor.getUniqueIdentifier();

		String targetHost = cam.getTargetHost();
		int targetPort = cam.getTargetPort();

		if (actor.getUAN() != null) {
			try {
				namingService.put( actor.getUAN(), new UAL("rmsp://" + targetHost + ":" + targetPort + actor.getUAN().getIdentifier()) );
			} catch (Exception e) {
				System.err.println("ERROR generating UAL for new actor: " + actor);
				System.err.println("Exception: " + e);
				e.printStackTrace();
				return;
			}
		}

		if (receptionService.isLocal(targetHost, targetPort)) {
			/********
				*	Actor will be created locally.
			 ********/
			actors.put(id, actor);
			Message constructMessage = actor.getConstructMessage();
			if (constructMessage != null) process(actor, constructMessage);
		} else {
			/********
				*	Actor will be created remotely.
			 ********/
			hosts.put(id, targetHost);
			ports.put(id, targetPort);
			transportService.transfer(actor, targetHost, targetPort);
		}
	}

	/********
		*	The following handles message lookup and invocation.
	 ********/
	public void process(Actor actor, Message message) {
		if (message.getRequiredTokens() > 0) {
			actor.handleRequiredTokens(message);
			return;
		}

		actor.setCurrentMessage(message);
		try {
			Object returnValue = actor.invokeMessage(message.getMethodName(), message.getArguments());
			if (!actor.isMigrating()) message.resolveContinuations(returnValue);
		} catch (CurrentContinuationException e) {
		} catch (Exception exception) {
			String exceptionMessage = "Message processing exception:\n";
			exceptionMessage += "\tSent by: " + message.getSource() + "\n";
			exceptionMessage += "\tReceived by actor: " + actor.toString() + "\n";
			exceptionMessage += "\tMessage: " + message.toString() + "\n";
			exceptionMessage += "\tThrew exception: " + exception + "\n";
			System.err.println(exceptionMessage);
			exception.printStackTrace();
			return;
		}
		Vector<Message> generatedMessages = actor.getGeneratedMessages();
		for (int i = 0; i < generatedMessages.size(); i++) {
			final Message current = generatedMessages.get(i);

			if (current.getProperty() != null) {
				if (current.getProperty().equals("delay")) {
					Object[] parameters = current.getPropertyParameters();
					final long sleepTime = ((Integer)parameters[0]).intValue();

					Runnable delayThread = new Runnable() {
						public void run() {
							try {
								Thread.sleep(sleepTime);
							} catch (Exception e) {
								System.err.println("Error Sending Message: ");
								System.err.println("\tError in delay thread caused by using the message delay property.");
								System.err.println("\tMessage: " + current.toString());
								System.err.println("\tException: " + e);
							}
							stageService.sendMessageFromLocal( current );
						}
					};
					new Thread(delayThread, "Delay Thread").start();
				} else {
					stageService.sendMessageFromLocal( current );
				}
			} else {
				stageService.sendMessageFromLocal( current );
			}
		}
		actor.setCurrentMessage(null);

		if (actor.isMigrating()) {
			String id = actor.getUniqueIdentifier();
			String migrationHost = actor.getMigrationHost();
			int migrationPort = actor.getMigrationPort();

			actors.remove(id);
			hosts.put(id, migrationHost);
			ports.put(id, migrationPort);			

			transportService.transfer(actor, migrationHost, migrationPort);
		} else if (actor.isDestroyed()) {
			String id = actor.getUniqueIdentifier();
			actors.remove(id);
			hosts.remove(id);
			ports.remove(id);
		}
	}

	private void sendUpdateMessage(SendFromRemoteMessage sfrm) {
		/********
			*	The actor is known about, but not local. update the reference and
			*	its source theater.
			*	1. check to see if a message has been sent to the source theater to
			*	   update this reference already
			*	2. if so, just send the message to it's target
			*	3. otherwise, create a message to send to the source theater to update
			*	   the reference.
			*
			*	id = sourceHost:sourcePort:invalidReferenceId
		 ********/
		String sourceHost = sfrm.getSourceHost();
		int sourcePort = sfrm.getSourcePort();

		String invalidReferenceId = sfrm.getMessage().getTargetIdentifier();

		String updateMessageId = sourceHost + ":" + sourcePort + ":" + invalidReferenceId;

		if (receptionService.isLocal(sourceHost, sourcePort)) {
			System.err.println("ERROR: sendUpdateMessage updating local theater!");
			System.err.println("\thost:port:: " + sourceHost + ":" + sourcePort + ", invalidReferenceId: " + invalidReferenceId);
			return;
		}

		if (alreadyUpdating.contains(updateMessageId)) {
			/********
				*	A message to update this reference has already been sent to the
				*	source theater, so do nothing.
			 ********/
			return;
		} else {
			alreadyUpdating.add(updateMessageId);
			transportService.transfer(new UpdateInvalidReferenceMessage(receptionService.getLocalHost(), receptionService.getLocalPort(), invalidReferenceId), sourceHost, sourcePort);
		}
	}

	private void handleUpdateInvalidReferenceMessage(UpdateInvalidReferenceMessage uirm) {
		String referenceId = uirm.getInvalidId();

		UAL ual = null;
		try {
			ual = namingService.get( new UAN(referenceId) );
		} catch (Exception e) {
			System.err.println("ERROR GETTING NEW REFERENCE ON UPDATE INVALID REFERENE: " + e);
			e.printStackTrace();
			return;
		}
		String correctHost = ual.getHost();
		int correctPort = ual.getPort();

		if (actors.get(referenceId) == null) {
			hosts.put(referenceId, correctHost);
			ports.put(referenceId, correctPort);
		}
		if (!receptionService.isLocal(correctHost, correctPort)) {
			Vector<Object> storedMessages = pendingMessages.remove(referenceId);
			if (storedMessages != null) {
				for (int i = 0; i < storedMessages.size(); i++) {
					Object current = storedMessages.get(i);
					if (current instanceof SendFromLocalMessage) put( (SendFromLocalMessage)current );
					else if (current instanceof SendFromRemoteMessage) put( (SendFromRemoteMessage)current );
					else put(current);
				}
			}
		}

		//send updated message
		transportService.transfer(new UpdatedInvalidReferenceMessage(ServiceFactory.getReception().getLocalHost(), ServiceFactory.getReception().getLocalPort(), referenceId), uirm.getSourceHost(), uirm.getSourcePort());
	}

	private void handleUpdatedInvalidReferenceMessage(UpdatedInvalidReferenceMessage uirm) {
		String updateMessageId = uirm.getSourceHost() + ":" + uirm.getSourcePort() + ":" + uirm.getInvalidId();
		alreadyUpdating.remove(updateMessageId);
	}

	public synchronized void put(ActorMigrationMessage amm) {
		mailbox.insertElementAt(amm, 0);
		notify();
	}

	public synchronized void put(CreateActorMessage cam) {
		mailbox.insertElementAt(cam, 0);
		notify();
	}

	public synchronized void put(Object object) {
		mailbox.add(object);
		notify();
	}

	public synchronized Object getNext() {
		if (mailbox.isEmpty()) {
			try {
				wait();
			} catch (Exception e) {
				System.err.println("LookupServiceActor error.  waiting on next message.");
				System.err.println("Exception: " + e);
				e.printStackTrace();
			}
		}
		return mailbox.remove(0);
	}
}
